#version 430

in vec3 gFacetNormal;
in vec2 gTexCoord;
in vec3 gEyeWS;
in vec3 gPosition;
in vec3 gPointLightDirWS;
in vec4 gShadowCoord;
in vec4 gAtt;


uniform vec3 directionalLight;
uniform sampler2D texHeight;
uniform vec3 LightPosition_worldspace;

uniform float isTorchOn;
uniform sampler2DShadow shadowMap;
uniform samplerCubeShadow shadowCube;

out vec4 fragColor;


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(gShadowCoord.xy + poissonDisk[i]/700.0, gShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}


float calculateVisibilityShadowCube()
{
	vec2 near_far = vec2(0.1f, 20.0f);
    vec4 lookUpVector = vec4((gPosition) - LightPosition_worldspace, 0);
	float currentDistance = length(lookUpVector) / (near_far.y);
	currentDistance = clamp(currentDistance, 0, 1);
	
	float angle = randomAngle(gPosition);
	float s = sin(angle);
	float c = cos(angle);
	
	float diskRadius = (1.0 + (1.0 - currentDistance) * 3) / (100);
	float compareDistance = currentDistance - 0.005;

	float shadowFactor = 0;
	float numSamples = 0;
    
	for(int i = 0; i < 8; ++i)
	{
		vec2 rotatedOffset = vec2(poissonDisk[i].x * c + poissonDisk[i].y * s, poissonDisk[i].x * -s + poissonDisk[i].y * c);
		vec3 lookup = lookUpVector.xyz + vec3(rotatedOffset * diskRadius, 0);
		lookup = normalize(lookup);
		shadowFactor += texture(shadowCube, vec4(lookup, compareDistance));
		numSamples++;
	}
	
	shadowFactor /= numSamples;
    return clamp(shadowFactor, 0, 1);
}




void main()
{
    vec3 color = vec3(1.0, 0.6, 0.1);
	vec3 ambient = color * 0.3;
	vec3 specColor = vec3(0.5, 0.5, 0.5);
	float bumpColor = texture(texHeight, gTexCoord).x;
	color = color * bumpColor;

	////////////////////////////////////
	// Shadow
	////////////////////////////////////
	float visD = calculateVisibilityShadowMap();
    float visP = calculateVisibilityShadowCube();
	float att = 1 - dot(gAtt/10.0f, gAtt/10.0f);
	att = max(att, 0.0f);

	////////////////////////////////////
	// Basic Lighting
	////////////////////////////////////
    vec3 N = normalize(gFacetNormal);

	vec3 LD = normalize(directionalLight);
	vec3 LP = normalize(gPointLightDirWS);

    float cosThetaD = clamp(dot(N, LD), 0, 1);
	float cosThetaP = clamp(dot(N, LP), 0, 1);

	float selfshadowD = clamp(1.2f * cosThetaD, 0, 1);
	float selfshadowP = clamp(1.2f * cosThetaP, 0, 1);

	float cosTheta =  (cosThetaP + cosThetaD) * 0.5f;
	float visibility = (selfshadowP * visP * att * isTorchOn + selfshadowD * visD ) * 0.5f;

	////////////////////////////////////
	// Specular Highlights
	////////////////////////////////////
	vec3 eye = normalize(gEyeWS);

	float intD = max(dot(N, LD), 0.0f);
	float intP = max(dot(N, LP), 0.0f);

    vec3 specD = vec3(0.3,0.3,0.3);
	vec3 specP = vec3(0.3,0.3,0.3);
	
    if(intD > 0.0) {
		vec3 h = normalize(LD + eye);
		float intSpec = max(dot(h, N), 0.0);
		specD = specColor * intSpec; 
	}

	if(intP > 0.0) {
		vec3 h = normalize(LP + eye);
		float intSpec = max(dot(h, N), 0.0);
		specP = specColor * intSpec; 
	}

    vec3 RD = reflect(-LD, N);
	vec3 RP = reflect(-LP, N);

    float cosAlphaD = clamp(dot(eye, RD), 0, 1);
	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
	float cosAlpha = (cosAlphaD + cosAlphaP) * 0.5;
	vec3 spec = (specD + specP) * 0.5;

	specColor = specColor * spec * pow(cosAlpha, 5);

    color = ambient + (visibility * color * cosTheta) + (specColor * bumpColor) * visibility;
	
    fragColor = vec4(color, 1.0);
}