#version 430

layout(triangles, equal_spacing, ccw) in;

in vec3 tcPosition[];
in vec3 tcNormal[];
in vec2 tcTexCoord[];

out vec3 tePosition;
out vec2 teTexCoord;
out vec3 teNormal;
out vec3 teEyeWS;
out vec3 tePointLightDirWS;
out vec4 teShadowCoord;
out vec4 teAtt;

uniform mat4 ModelViewProjection;
uniform mat4 M;
uniform mat4 DepthBiasMVP;
uniform mat3 N;
uniform vec3 cameraPositionWS;

uniform vec3 LightPosition_worldspace;

uniform sampler2D texHeight;

void main()
{
    vec3 p0 = gl_TessCoord.x * tcPosition[0];
    vec3 p1 = gl_TessCoord.y * tcPosition[1];
    vec3 p2 = gl_TessCoord.z * tcPosition[2];
    vec3 pos = (p0 + p1 + p2);

    vec3 n0 = gl_TessCoord.x * tcNormal[0];
    vec3 n1 = gl_TessCoord.y * tcNormal[1];
    vec3 n2 = gl_TessCoord.z * tcNormal[2];
    vec3 normal = normalize(n0 + n1 + n2);

    vec2 tc0 = gl_TessCoord.x * tcTexCoord[0];
    vec2 tc1 = gl_TessCoord.y * tcTexCoord[1];
    vec2 tc2 = gl_TessCoord.z * tcTexCoord[2];  
    teTexCoord = (tc0 + tc1 + tc2);

    float height = texture(texHeight, teTexCoord).x;
    pos += normal * (height * 0.2f);
	pos = (pos);
	
    gl_Position = ModelViewProjection * vec4(pos, 1);

	vec3 tmpPos = vec3(M * vec4(pos, 1.0)).xyz;

    teNormal    = N * normal;
    tePosition  = tmpPos;
	teEyeWS		= (cameraPositionWS - tmpPos);
	tePointLightDirWS = (LightPosition_worldspace - tmpPos);
	teShadowCoord = DepthBiasMVP * vec4(pos, 1.0);
	teAtt = vec4(tePointLightDirWS, 0) * 0.4f;
}