#version 330 core

// Input vertex data, different for all executions of this shader.
layout(location = 0) in vec3 vertexPosition_modelspace;
layout(location = 1) in vec2 vertexUV;
layout(location = 2) in vec3 vertexNormal_modelspace;
layout(location = 3) in vec3 vertexTangent_modelspace;
layout(location = 4) in vec3 vertexBitangent_modelspace;

// Output data ; will be interpolated for each fragment.
out vec2 UV;
out vec3 wsPosition;
out vec4 ShadowCoord;
out vec4 attenuation;
out vec3 tsNormal;
out vec3 tsEyeDirection;
out vec3 tsLightDirection;

// Values that stay constant for the whole mesh.
uniform mat4 MVP;
uniform mat4 M;
uniform mat4 DepthBiasMVP;
uniform vec3 invLightDirection;
uniform vec3 wsCameraPosition;
uniform float invBumpDepth;


void main()
{

	// Output position of the vertex, in clip space : MVP * position
	gl_Position =  MVP * vec4(vertexPosition_modelspace,1);
	
    ShadowCoord = DepthBiasMVP * vec4(vertexPosition_modelspace,1);
    attenuation = vec4(0);
    
	// Position of the vertex, in worldspace : M * position
	wsPosition = (M * vec4(vertexPosition_modelspace,1)).xyz;
	
	// UV of the vertex. No special space for this one.
	UV = vertexUV;


	///////////////////////////////////////////////////////////////////////
	// Rest ... in Texture-Space --> required for Displacement mapping

	vec3 vertexTangent_cameraspace = (M * vec4(vertexTangent_modelspace, 0.0)).xyz;
	vec3 vertexBitangent_cameraspace = (M * vec4(vertexBitangent_modelspace, 0.0)).xyz;
	vec3 vertexNormal_cameraspace = (M * vec4(vertexNormal_modelspace, 0.0)).xyz;

	mat3 TBN = mat3(normalize(vertexTangent_cameraspace), normalize(vertexBitangent_cameraspace), normalize(vertexNormal_cameraspace));
	
	vec3 EyeDirection = TBN * (wsCameraPosition - ( M * vec4(vertexPosition_modelspace,1)).xyz);
	EyeDirection.z *= -invBumpDepth;
	tsEyeDirection = EyeDirection;

	// Normal of the the vertex, in camera space
	tsNormal = TBN * (M * vec4(vertexNormal_modelspace,0)).xyz; // Only correct if ModelMatrix does 
	
	// Light Direction
	tsLightDirection = TBN * (invLightDirection).xyz;
	///////////////////////////////////////////////////////////////////////
}