#version 330 core 

in vec2 noise_uv1;
in vec2 noise_uv2;
in vec2 noise_uv3;
in vec2 UV;

out vec4 color;

uniform sampler2D fireTexture;
uniform sampler2D alphaTexture;
uniform sampler2D noiseTexture;
uniform float distortionScale;
uniform float distortionBias;
uniform vec2 distortion1;
uniform vec2 distortion2;
uniform vec2 distortion3;

void main()
{
	vec4 no1 = texture(noiseTexture, noise_uv1);
	vec4 no2 = texture(noiseTexture, noise_uv2);
	vec4 no3 = texture(noiseTexture, noise_uv3);
	
	no1 = no1 * 2.0f - 1.0f;
	no2 = no2 * 2.0f - 1.0f;
	no3 = no3 * 2.0f - 1.0f;
	
	no1.xy = no1.xy * distortion1;
	no2.xy = no2.xy * distortion2;
	no3.xy = no3.xy * distortion3;
	
	vec4 finalNoise = no1 + no2 + no3;
	
	float perturb = (UV.y) * distortionScale + distortionBias;
	
	vec2 noiseCoord = finalNoise.xy * perturb + UV;
	
	vec4 fireColor = texture(fireTexture, noiseCoord);
	vec4 alphaColor = texture(alphaTexture, noiseCoord);
	
	fireColor.a = alphaColor.r / 10;
	
	color = fireColor;
} 