#version 430 core

// Input vertex data, different for all executions of this shader.
layout(location = 0) in vec3 vertexPosition_modelspace;
layout(location = 1) in vec2 vertexUV;
layout(location = 2) in vec3 vertexNormal_modelspace;
layout(location = 3) in vec3 vertexTangent_modelspace;
layout(location = 4) in vec3 vertexBitangent_modelspace;

// Output data ; will be interpolated for each fragment.
out mat3 TBN;
out vec4 attenuation;
out vec4 ShadowCoord;
out vec3 Position_worldspace;
out vec3 Normal_worldspace;
out vec3 EyeDirection_worldspace;
out vec3 LightDirection_worldspace;
out vec3 LightDirection_worldspace_directional;
out vec2 UV;


// Values that stay constant for the whole mesh.
uniform mat4 MVP; // Model-View-Projection
uniform mat4 V; // View
uniform mat4 M; // Model
uniform mat4 ITM; // Inverse Transpose of Model
uniform mat4 DepthBiasMVP;
uniform vec3 LightInvDirection_worldspace;
uniform vec3 LightPosition_worldspace;
uniform vec3 CameraPosition_worldspace;
uniform float lightRange;

void main()
{

	// Output position of the vertex, in clip space : MVP * position
	gl_Position =  MVP * vec4(vertexPosition_modelspace,1);
    
    // Required for sampling shadow texture
    ShadowCoord = DepthBiasMVP * vec4(vertexPosition_modelspace,1);
    
	// Position of the vertex, in worldspace : M * position
	Position_worldspace = (M * vec4(vertexPosition_modelspace,1)).xyz;
	
	vec3 vertexTangent_ws = (M * vec4(vertexTangent_modelspace, 0.0)).xyz;
	vec3 vertexBitangent_ws = (M * vec4(vertexBitangent_modelspace, 0.0)).xyz;
	vec3 vertexNormal_ws = (M * vec4(vertexNormal_modelspace, 0.0)).xyz;

	TBN = mat3(normalize(vertexTangent_ws), normalize(vertexBitangent_ws), normalize(vertexNormal_ws));
	
	vec3 EyeDirection = (CameraPosition_worldspace - ( M * vec4(vertexPosition_modelspace,1)).xyz);
	EyeDirection_worldspace = EyeDirection;
	
	// Normal of the the vertex, in camera space
	Normal_worldspace = (ITM * vec4(vertexNormal_modelspace,0)).xyz;
	
	// UV of the vertex. No special space for this one.
	UV = vertexUV;
	
    LightDirection_worldspace_directional = LightInvDirection_worldspace.xyz; //(V * vec4(LightInvDirection_worldspace,0)).xyz; //  
    
	// Vector that goes from the vertex to the light, in camera space
	LightDirection_worldspace = ((M * vec4(LightPosition_worldspace, 1) - (M * vec4(vertexPosition_modelspace,1)))).xyz;
	
	attenuation = vec4(LightDirection_worldspace, 0) * lightRange;
	
}