#version 430 core

in mat3 TBN;
in vec3 LightDirection_worldspace_point;
in vec3 LightDirection_worldspace_directional;
in vec3 EyeDirection_worldspace;
in vec3 Normal_worldspace;
in vec2 UV;
in vec2 sUV;
in vec4 ShadowCoord;

in vec3 spotVertexVector1;
in vec3 spotVertexVector2;
in vec3 spotVertexVector3;
in vec3 spotColor1;
in vec3 spotColor2;
in vec3 spotColor3;
in vec3 spotPos1;
in vec3 spotPos2;
in vec3 spotPos3;

/*********************************************
 * for shadow-cube
 *********************************************/
in vec3 Position_worldspace;

uniform vec3 LightPosition_worldspace;
uniform vec2 near_far;
/*********************************************/

uniform sampler2D MainTextureSampler;

uniform sampler2D Ground1TextureSampler;
uniform sampler2D Ground2TextureSampler;
uniform sampler2D StoneTextureSampler;
uniform sampler2D SandTextureSampler;

uniform sampler2D Ground1TextureSamplerSpec;
uniform sampler2D Ground2TextureSamplerSpec;
uniform sampler2D StoneTextureSamplerSpec;
uniform sampler2D SandTextureSamplerSpec;

uniform sampler2D Ground1TextureSamplerNorm;
uniform sampler2D Ground2TextureSamplerNorm;
uniform sampler2D StoneTextureSamplerNorm;
uniform sampler2D SandTextureSamplerNorm;

uniform float shadowThreshold;
uniform float ambientBrightness;

uniform float isTorchOn;

uniform sampler2DShadow shadowMap; // shadowMap
uniform samplerCubeShadow shadowCube; // shadowCube

uniform float normalMappingOnOff;

out vec3 color;


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 diffColor;
vec3 ambientColor;
vec3 specColor;
vec3 textureNormal;
/*********************************************************
 *********************************************************/



float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float calculateVisibilityShadowCube()
{
    vec4 lookUpVector = vec4((Position_worldspace) - LightPosition_worldspace, 0);
	float currentDistance = length(lookUpVector) / (near_far.y);
	currentDistance = clamp(currentDistance, 0, 1);
	
	float angle = randomAngle(Position_worldspace);
	float s = sin(angle);
	float c = cos(angle);
	
	float diskRadius = (1.0 + (1.0 - currentDistance) * 3) / (100);
	float compareDistance = currentDistance - 0.005;

	float shadowFactor = 0;
	float numSamples = 0;
    
	for(int i = 0; i < 8; ++i)
	{
		vec2 rotatedOffset = vec2(poissonDisk[i].x * c + poissonDisk[i].y * s, poissonDisk[i].x * -s + poissonDisk[i].y * c);
		vec3 lookup = lookUpVector.xyz + vec3(rotatedOffset * diskRadius, 0);
		lookup = normalize(lookup);
		shadowFactor += texture(shadowCube, vec4(lookup, compareDistance));
		numSamples++;
	}
	
	shadowFactor /= numSamples;
    return clamp(shadowFactor, 0, 1);
}



bool areSimilar(vec3 a, vec3 b)
{
	float c1 = abs(a.x - b.x);
	float c2 = abs(a.y - b.y);
	float c3 = abs(a.z - b.z);
	return (c1 <= 0.1) && (c2 <= 0.1) && (c3 <= 0.1);
}


void getColor()
{
	vec3 cLand  = vec3(  0.0f, 127.0f / 255.0f, 16.0f / 255.0f);
	vec3 cSand  = vec3(  1.0f, 245.0f / 255.0f, 191.0f / 255.0f);
	vec3 cRock  = vec3(131.0f / 255.0f, 131.0f / 255.0f, 131.0f / 255.0f);
	vec3 cTree  = vec3(  0.0f,  51.0f / 255.0f,  16.0f / 255.0f);

	vec3 compColor = texture(MainTextureSampler, UV).rgb;
	vec3 tmpNormal = vec3(0,0,0);
	vec3 tmpSpec = vec3(0,0,0);


	if (areSimilar(compColor, cRock)) {
		diffColor = texture(StoneTextureSampler, sUV).rgb;
		tmpNormal = texture(StoneTextureSamplerNorm, sUV).rgb;
		tmpSpec   = texture(StoneTextureSamplerSpec, sUV).rgb;
	}
	else if (areSimilar(compColor, cLand)) {
		diffColor = texture(Ground1TextureSampler, sUV).rgb;
		tmpNormal = texture(Ground1TextureSamplerNorm, sUV).rgb;
		tmpSpec   = texture(Ground1TextureSamplerSpec, sUV).rgb;
	}
	else if (areSimilar(compColor, cSand)) {
		diffColor = texture(SandTextureSampler, sUV).rgb;
		tmpNormal = texture(SandTextureSamplerNorm, sUV).rgb;
		tmpSpec   = texture(SandTextureSamplerSpec, sUV).rgb;
	}
	else {
		diffColor = texture(Ground2TextureSampler, sUV).rgb;
		tmpNormal = texture(Ground2TextureSamplerNorm, sUV).rgb;
		tmpSpec   = texture(Ground2TextureSamplerSpec, sUV).rgb;
	}

	textureNormal = normalize(tmpNormal) * 2.0 - vec3(1.0);
	specColor = tmpSpec;
	ambientColor = diffColor * ambientBrightness;
}


void getNormalSpecularMapping()
{
	/*********************************************************
	 * Shadow
	 *********************************************************/
	float visD = calculateVisibilityShadowMap();
    float visP = calculateVisibilityShadowCube();

	/*********************************************************
	 * Attenuation
	 *********************************************************/
	float att = 1 - dot(LightDirection_worldspace_point.xyz/40.0f, LightDirection_worldspace_point.xyz/40.0f);
	att = max(att, 0.0f);

	float attD = 1.0f;

	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float adjustBumpValue = 2.0f;

	vec3 n =  normalize(transpose(TBN) * (adjustBumpValue * textureNormal) + Normal_worldspace) * normalMappingOnOff + (1.0f - normalMappingOnOff) * normalize(Normal_worldspace);

	vec3 lP = normalize(LightDirection_worldspace_point);
	vec3 lD = normalize(LightDirection_worldspace_directional);

	float cosThetaP = clamp(dot(n, lP), 0, 1);
	float cosThetaD = clamp(dot(n, lD), 0, 1);

	float selfshadowP = clamp(shadowThreshold * cosThetaP, 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);
	
	/*********************************************************
	 * Spots
	 *********************************************************/
	
	vec3 lS1 = normalize(spotVertexVector1);
	vec3 lS2 = normalize(spotVertexVector2);
	vec3 lS3 = normalize(spotVertexVector3);

	float att1 = max((1 - dot(spotVertexVector1.xyz/40.0f, spotVertexVector1.xyz/40.0f)), 0.0f);
	float att2 = max((1 - dot(spotVertexVector2.xyz/40.0f, spotVertexVector2.xyz/40.0f)), 0.0f);
	float att3 = max((1 - dot(spotVertexVector3.xyz/40.0f, spotVertexVector3.xyz/40.0f)), 0.0f);
	
	float cosThetaS1 = clamp(dot(n, lS1), 0, 1);
	float cosThetaS2 = clamp(dot(n, lS2), 0, 1);
	float cosThetaS3 = clamp(dot(n, lS3), 0, 1);
	
	float selfshadowS1 = clamp(shadowThreshold * cosThetaS1, 0, 1);
	float selfshadowS2 = clamp(shadowThreshold * cosThetaS2, 0, 1);
	float selfshadowS3 = clamp(shadowThreshold * cosThetaS3, 0, 1);

	float v1 = 1.0f;
	float v2 = 1.0f;
	float v3 = 1.0f;

	float angle = 0.8f;

	vec3 sp = vec3(0, 1, 0);
	if(dot(sp, lS1) < angle){
		v1 = 0.0f;
	}
	if(dot(sp, lS2) < angle){
		v2 = 0.0f;
	}
	if(dot(sp, lS3) < angle){
		v3 = 0.0f;
	}

	vec3 spotComponent = selfshadowS1 * v1 * spotColor1 * att1 + 
						 selfshadowS2 * v2 * spotColor2 * att2 +
						 selfshadowS3 * v3 * spotColor3 * att3;
	spotComponent = spotComponent * 0.4f;


	/*********************************************************
	 * Specular Mapping
	 *********************************************************/
	vec3 eye = normalize(EyeDirection_worldspace);

	float intensityP = max(dot(n, lP), 0.0f);
	float intensityD = max(dot(n, lD), 0.0f);

	vec3 base = vec3(0.5, 0.5, 0.5);
	vec3 specP = base;
	vec3 specD = base;


	if(intensityP > 0.0) {
		vec3 hP = normalize(lP + eye);
		float intSpecP = max(dot(hP, n), 0.0);
		specP = specColor * intSpecP;
	}
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = specColor * intSpecD; 
	}


	vec3 RP = reflect(-lP, n);
    vec3 RD = reflect(-lD, n);

	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);


	/*********************************************************
	 * Final Results
	 *********************************************************/
	float visibility = (selfshadowP * visP * att * isTorchOn + selfshadowD * visD * attD) * 0.5;
	float cosTheta =  (cosThetaP + cosThetaD) * 0.5;
	float cosAlpha =  (cosAlphaP + cosAlphaD) * 0.5;

	vec3 spec =  (specP * 4.0f + specD * 4.0f) * 0.125f;

	ambientColor = (ambientColor + spotComponent);
	diffColor = (diffColor) * cosTheta * visibility;
	specColor = specColor * spec * pow(cosAlpha, 5) * visibility;
}

void main()
{
	getColor();
	getNormalSpecularMapping();

	color = ambientColor + diffColor + specColor;
}