#version 430 core

in mat3 TBN;
in vec3 LightDirection_worldspace_point;
in vec3 LightDirection_worldspace_directional;
in vec3 EyeDirection_worldspace;
in vec3 Normal_worldspace;
in vec2 UV;
in vec4 ShadowCoord;
in vec4 attenuation;
in vec4 attenuationD;

uniform sampler2D TextureSamplerDiff;
uniform sampler2D TextureSamplerSpec;
uniform sampler2D TextureSamplerNorm;

uniform float shadowThreshold;
uniform float ambientBrightness;

uniform sampler2DShadow shadowMap;

uniform float isTorchOn;

out vec3 color;


/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 diffColor;
vec3 ambientColor;
vec3 specColor;
vec3 textureNormal;

vec3 spec;
float cosAlpha;
float cosTheta;
float selfShadow;
/*********************************************************
 *********************************************************/


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}


void getColor()
{
	textureNormal = normalize(texture(TextureSamplerNorm, UV).rgb * 2.0 - 1.0);
	diffColor = texture(TextureSamplerDiff, UV).rgb;
	specColor = texture(TextureSamplerSpec, UV).rgb * 0.9;

	ambientColor = diffColor * (ambientBrightness - 0.05);
}


void getNormalSpecularMapping()
{
	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float visD = calculateVisibilityShadowMap();

	/************************* Attenuation *************************/
	float att = 1 - dot(attenuation/10.0f, attenuation/10.0f);
	att = max(att, 0.0f);

	float attD = 1 - dot(attenuationD/10.0f, attenuationD/10.0f);
	attD = max(attD, 0.0f);

	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float adjustBumpValue = 0.5f;

	vec3 n =  normalize(transpose(TBN) * (adjustBumpValue * textureNormal) + Normal_worldspace);

	vec3 lP = normalize(LightDirection_worldspace_point);
	vec3 lD = normalize(LightDirection_worldspace_directional);
    
	float cosThetaP = clamp(dot(n, lP), 0, 1);
	float cosThetaD = clamp(dot(n, lD), 0, 1);

    float selfshadowP = clamp(shadowThreshold * cosThetaP, 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);

	/*********************************************************
	 * Specular Mapping
	 *********************************************************/
	vec3 eye = normalize(EyeDirection_worldspace);

	float intensityP = max(dot(n, lP), 0.0f);
	float intensityD = max(dot(n, lD), 0.0f);
	vec3 specP = vec3(0.3,0.3,0.3);
	vec3 specD = vec3(0.3,0.3,0.3);

	if(intensityP > 0.0) {
		vec3 hP = normalize(lP + eye);
		float intSpecP = max(dot(hP, n), 0.0);
		specP = specColor * intSpecP;
	}
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = specColor * intSpecD; 
	}

	vec3 RP = reflect(-lP, n);
    vec3 RD = reflect(-lD, n);
	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);

	/*********************************************************
	 * Final Results
	 *********************************************************/

	selfShadow = (selfshadowP * 5.0f * att * isTorchOn + selfshadowD * 3.0f * visD * attD) * 0.125f;
	cosTheta =  (cosThetaP * 5.0f + cosThetaD * 3.0f) * 0.125f;
	cosAlpha =  (cosAlphaP * 5.0f + cosAlphaD * 3.0f) * 0.125f;
	spec =  (specP * 5.0f + specD * 3.0f) * 0.125f;
}

void main()
{
	getColor();
	getNormalSpecularMapping();

	color = ambientColor + (diffColor * cosTheta + specColor * spec * pow(cosAlpha, 5)) * selfShadow;
}