#version 430 core

in vec2 sUV;
in vec3 EyeCS;
in vec3 LightCS;
in vec3 tangent;
in vec3 bitangent;
in vec4 ShadowCoord;

uniform sampler2D WaterTextureSampler;
uniform sampler2D WaterTextureSamplerSpec;
uniform sampler2D WaterTextureSamplerNorm;

uniform float waveTime;
uniform float waveHeight;
uniform float waveWidth;
uniform mat4 M;

uniform float shadowThreshold;
uniform float ambientBrightness;

uniform sampler2DShadow shadowMap;

uniform float normalMappingOnOff;

out vec3 color;

vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 diffColor;
vec3 ambientColor;
vec3 specColor;
vec3 textureNormal;
/*********************************************************
 *********************************************************/

void getColor()
{
	diffColor = texture(WaterTextureSampler, sUV).rgb;
	textureNormal = normalize(texture(WaterTextureSamplerNorm, sUV).rgb * 2.0 - 1.0);
	specColor = texture(WaterTextureSamplerSpec, sUV).rgb;
	ambientColor = diffColor * ambientBrightness;
}

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}


void getNormalSpecularMapping()
{
	/*********************************************************
	 * Shadow Mapping
	 *********************************************************/
	float visD = calculateVisibilityShadowMap();

	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float adjustBumpValue = 0.5f;

	vec3 tmpNormal = cross(tangent, bitangent);

	vec3 newtangent = tangent;
	if(dot(cross(tmpNormal, newtangent), bitangent) < 0.0f)
	{
		newtangent = newtangent * -1.0f;
	}
	
	vec3 vTang = (M * vec4(newtangent, 0.0)).xyz;
	vec3 vBitang = (M * vec4(bitangent, 0.0)).xyz;
	vec3 vNormal = (M * vec4(tmpNormal, 0.0)).xyz;
	
	mat3 TBN = (mat3(normalize(vTang), normalize(vBitang), normalize(vNormal)));
	/******************************************************/
	vec3 n = normalize(transpose(TBN) * textureNormal) * normalMappingOnOff + (1.0f - normalMappingOnOff) * normalize(vNormal);
	/******************************************************/

	vec3 lD = normalize(LightCS);
	float cosThetaD = clamp(dot(n, lD), 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);

	/*********************************************************
	 * Specular Mapping
	 *********************************************************/
	vec3 eye = normalize(EyeCS);

	float intensityD = max(dot(n, lD), 0.0f);
	vec3 specD = vec3(0.3,0.3,0.3);
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = specColor * intSpecD; 
	}

    vec3 RD = reflect(-lD, n);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);

	/*********************************************************
	 * Final Results
	 *********************************************************/
	float adjust = 0.3f;
	float visibility = (selfshadowD) * visD;
	float cosTheta	 = (cosThetaD);
	float cosAlpha	 = (cosAlphaD);
	vec3 spec		 = (specD) * adjust;

	diffColor = diffColor * cosTheta * visibility;
	specColor = specColor * spec * pow(cosAlpha, 5) * visibility;
}



void main()
{
	getColor();
	getNormalSpecularMapping();
	color = ambientColor + diffColor + specColor;
}