#version 430 core

in mat3 TBN;
in vec3 LightDirection_worldspace_point;
in vec3 LightDirection_worldspace_directional;
in vec3 EyeDirection_worldspace;
in vec3 Normal_worldspace;
in vec2 UV;
in vec4 ShadowCoord;
in vec3 Position_worldspace;
in vec3 spotVertexVector1;
in vec3 spotVertexVector2;
in vec3 spotVertexVector3;
in vec3 spotColor1;
in vec3 spotColor2;
in vec3 spotColor3;
in vec3 spotPos1;
in vec3 spotPos2;
in vec3 spotPos3;


uniform vec3 LightPosition_worldspace;

uniform sampler2D TextureSamplerDiff;
uniform sampler2D TextureSamplerSpec;
uniform sampler2D TextureSamplerNorm;

uniform float shadowThreshold;
uniform float ambientBrightness;
uniform float isTorchOn;
uniform float normalMappingOnOff;

uniform int UseShadowCube;
uniform int UseSpots;

uniform sampler2DShadow shadowMap;
uniform samplerCubeShadow shadowCube;



out vec3 color;

vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 diffColor;
vec3 ambientColor;
vec3 specColor;
vec3 textureNormal;

/*********************************************************
 *********************************************************/

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float calculateVisibilityShadowCube()
{
    vec4 lookUpVector = vec4((Position_worldspace) - LightPosition_worldspace, 0);
	float currentDistance = length(lookUpVector) / (20.0f);
	currentDistance = clamp(currentDistance, 0, 1);
	
	float angle = randomAngle(Position_worldspace);
	float s = sin(angle);
	float c = cos(angle);
	
	float diskRadius = (1.0 + (1.0 - currentDistance) * 3) / (100);
	float compareDistance = currentDistance - 0.005;

	float shadowFactor = 0;
	float numSamples = 0;
    
	for(int i = 0; i < 8; ++i)
	{
		vec2 rotatedOffset = vec2(poissonDisk[i].x * c + poissonDisk[i].y * s, poissonDisk[i].x * -s + poissonDisk[i].y * c);
		vec3 lookup = lookUpVector.xyz + vec3(rotatedOffset * diskRadius, 0);
		lookup = normalize(lookup);
		shadowFactor += texture(shadowCube, vec4(lookup, compareDistance));
		numSamples++;
	}
	
	shadowFactor /= numSamples;
    return clamp(shadowFactor, 0, 1);
}

vec3 getSpot(vec3 n)
{
	vec3 lS1 = normalize(spotVertexVector1);
	vec3 lS2 = normalize(spotVertexVector2);
	vec3 lS3 = normalize(spotVertexVector3);

	float att1 = max((1 - dot(spotVertexVector1.xyz/40.0f, spotVertexVector1.xyz/40.0f)), 0.0f);
	float att2 = max((1 - dot(spotVertexVector2.xyz/40.0f, spotVertexVector2.xyz/40.0f)), 0.0f);
	float att3 = max((1 - dot(spotVertexVector3.xyz/40.0f, spotVertexVector3.xyz/40.0f)), 0.0f);
	
	float cosThetaS1 = clamp(dot(n, lS1), 0, 1);
	float cosThetaS2 = clamp(dot(n, lS2), 0, 1);
	float cosThetaS3 = clamp(dot(n, lS3), 0, 1);
	
	float selfshadowS1 = clamp(shadowThreshold * cosThetaS1, 0, 1);
	float selfshadowS2 = clamp(shadowThreshold * cosThetaS2, 0, 1);
	float selfshadowS3 = clamp(shadowThreshold * cosThetaS3, 0, 1);

	float v1 = 1.0f;
	float v2 = 1.0f;
	float v3 = 1.0f;

	float angle = 0.8f;

	vec3 sp = vec3(0, 1, 0);
	if(dot(sp, lS1) < angle){
		v1 = 0.0f;
	}
	if(dot(sp, lS2) < angle){
		v2 = 0.0f;
	}
	if(dot(sp, lS3) < angle){
		v3 = 0.0f;
	}

	vec3 spotComponent = selfshadowS1 * v1 * spotColor1 * att1 + 
						 selfshadowS2 * v2 * spotColor2 * att2 +
						 selfshadowS3 * v3 * spotColor3 * att3;
	spotComponent = spotComponent * 0.4f;
	return spotComponent;
}


void getColor()
{
	textureNormal = normalize(texture(TextureSamplerNorm, UV).rgb * 2.0 - 1.0);
	diffColor = texture(TextureSamplerDiff, UV).rgb;
	specColor = texture(TextureSamplerSpec, UV).rgb * 0.9;
	ambientColor = diffColor * ambientBrightness;
}


void getNormalSpecularMapping()
{
	/*********************************************************
	 * Shadow Mapping
	 *********************************************************/
	float visD = calculateVisibilityShadowMap();
	float visP = 1.0f;
	if(UseShadowCube == 1) {
		visP = calculateVisibilityShadowCube();
	}

	/************************* Attenuation *************************/
	float att = max((1 - dot(LightDirection_worldspace_point.xyz/40.0f, LightDirection_worldspace_point.xyz/40.0f)), 0.0f);
	float attD = 1.0f;

	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float adjustBumpValue = 0.5f;

	vec3 n =  normalize(transpose(TBN) * (adjustBumpValue * textureNormal) + Normal_worldspace) * normalMappingOnOff + (1.0f - normalMappingOnOff) * normalize(Normal_worldspace);

	vec3 lP = normalize(LightDirection_worldspace_point);
	vec3 lD = normalize(LightDirection_worldspace_directional);
    
	float cosThetaP = clamp(dot(n, lP), 0, 1);
	float cosThetaD = clamp(dot(n, lD), 0, 1);

    float selfshadowP = clamp(shadowThreshold * cosThetaP, 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);

	/*********************************************************
	 * Specular Mapping
	 *********************************************************/
	vec3 eye = normalize(EyeDirection_worldspace);

	float intensityP = max(dot(n, lP), 0.0f);
	float intensityD = max(dot(n, lD), 0.0f);
	vec3 specP = vec3(0.3,0.3,0.3);
	vec3 specD = vec3(0.3,0.3,0.3);

	if(intensityP > 0.0) {
		vec3 hP = normalize(lP + eye);
		float intSpecP = max(dot(hP, n), 0.0);
		specP = specColor * intSpecP;
	}
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = specColor * intSpecD; 
	}

	vec3 RP = reflect(-lP, n);
    vec3 RD = reflect(-lD, n);
	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);

	/*********************************************************
	 * Final Results
	 *********************************************************/
	float visibility = (selfshadowP * att * visP * isTorchOn + selfshadowD * visD * attD) * 0.5f;
	float cosTheta =  (cosThetaP + cosThetaD) * 0.5f;
	float cosAlpha =  (cosAlphaP + cosAlphaD) * 0.5f;
	vec3 spec =  (specP + specD) * 0.5f;

	if(UseSpots == 1) {
		ambientColor = ambientColor + getSpot(n);
	}

	diffColor = diffColor * cosTheta * visibility;
	specColor = specColor * spec * pow(cosAlpha, 5) * visibility;
}



void main()
{
	getColor();
	getNormalSpecularMapping();
	color = ambientColor + diffColor + specColor;
}