#version 430 core
layout(location = 0) in vec3 vertexPosition_modelspace;
layout(location = 1) in vec2 vertexUV;
layout(location = 2) in vec3 normals_modelspace;
layout(location = 3) in vec3 tangents_modelspace;
layout(location = 4) in vec3 bitangents_modelspace;

out mat3 TBN;
out vec4 ShadowCoord;
out vec3 LightDirection_worldspace_point;
out vec3 LightDirection_worldspace_directional;
out vec3 EyeDirection_worldspace;
out vec3 Normal_worldspace;
out vec3 Position_worldspace;
out vec2 UV;

out vec3 spotVertexVector1;
out vec3 spotVertexVector2;
out vec3 spotVertexVector3;
out vec3 spotColor1;
out vec3 spotColor2;
out vec3 spotColor3;
out vec3 spotPos1;
out vec3 spotPos2;
out vec3 spotPos3;

uniform mat4 MVP;
uniform mat4 M;
uniform mat4 V;
uniform mat4 ITM;
uniform mat4 DepthBiasMVP;
uniform vec3 LightInvDirection_worldspace;
uniform vec3 CameraPosition_worldspace;
uniform vec3 LightPosition_worldspace;
uniform vec3[90] offsets;

uniform vec3[3] spotpos;
uniform vec3[3] spotcolor;

void main()
{
	vec3 offset = offsets[gl_InstanceID];
	vec3 pos = vertexPosition_modelspace + offset;
	Position_worldspace = (M * vec4(pos, 1)).xyz;
    gl_Position = MVP * vec4(pos, 1.0f);

	ShadowCoord = DepthBiasMVP * vec4(pos, 1);

	vec3 vertexNormal_worldspace = (M * vec4(normals_modelspace, 0.0)).xyz;
	vec3 vertexTangent_worldspace = (M * vec4(tangents_modelspace, 0.0)).xyz;
	vec3 vertexBitangent_worldspace = (M * vec4(bitangents_modelspace, 0.0)).xyz;
	
	TBN = mat3(normalize(vertexTangent_worldspace), normalize(vertexBitangent_worldspace), normalize(vertexNormal_worldspace));
	Normal_worldspace = (ITM * vec4(vertexNormal_worldspace, 0)).xyz;
	EyeDirection_worldspace = (CameraPosition_worldspace - ( M * vec4(pos, 1)).xyz);

	LightDirection_worldspace_point = ((vec4(LightPosition_worldspace, 1) - (M * vec4(pos, 1)))).xyz;
	LightDirection_worldspace_directional = LightInvDirection_worldspace.xyz;

	UV = vertexUV;


	spotPos1 = spotpos[0];
	spotPos2 = spotpos[1];
	spotPos3 = spotpos[2];
	spotVertexVector1 = (spotPos1 - Position_worldspace);
	spotVertexVector2 = (spotPos2 - Position_worldspace);
	spotVertexVector3 = (spotPos3 - Position_worldspace);

	spotColor1 = spotcolor[0];
	spotColor2 = spotcolor[1];
	spotColor3 = spotcolor[2];

}
